#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <stdbool.h>
#include <string.h>
#include <AL/al.h>
#include <AL/alc.h>
#include <sbc/sbc.h>
#include <openaptx.h>

#ifdef __EMSCRIPTEN__
#include <emscripten.h>
#endif

// Buffer size of a single audio buffer. For some reason, there are cracklings with _some_ bitrate configurations
// if you set high buffer size.
#define BUFSIZE 8*1024
// Number of buffers of BUFSIZE.
#define NUM_BUFFERS 12
// Global sample rate for OpenAL and SBC encoder
#define SAMPLE_RATE 44100

#if SAMPLE_RATE == 44100
#define SBC_SAMPLE_RATE SBC_FREQ_44100
#elif SAMPLE_RATE == 48000
#define SBC_SAMPLE_RATE SBC_FREQ_48000
#endif

// Emscripten 'fps' - number of main loop calls per second
#define EMSCRIPTEN_FPS 12

#define WAIT_MORE 2
// Number of 'process' function calls to wait when the playback has ended.
// This is needed to play all pending sound buffers.
#define CLOSE_TICK_WAIT 3
int close_tick = 0;

#define CODEC_SBC 1
#define CODEC_APTX 2
#define CODEC_APTX_HD 3
// Default codec
int current_codec = CODEC_SBC;
//int current_codec = CODEC_APTX;
char *codec_names[] = {"NONE", "SBC", "aptX", "aptX HD"};
char *mode_names[] = {"MONO", "DUAL CHANNEL", "STEREO", "JOINT STEREO"};

ALCcontext *context;
ALCdevice *device;
ALuint source;
ALuint buffer[NUM_BUFFERS];
ALint buffer_queue, buffer_queue_i;
bool openal_initialized = false;

FILE *fp = NULL;
/* SBC supports very high bitrates, up to 5688 kbit/s (much
 * more than original 44100 Hz 16 bit audio stream!)
 * BUFSIZE*4 should be sufficient. */
char audiobuffer[BUFSIZE], enc_output_audiobuffer[BUFSIZE*4];

// Internal buffer for aptX 16→24 bit conversion
unsigned char int_buffer[BUFSIZE*2];
size_t fr;

// SBC encoder and decoder
sbc_t sbc_enc = {0}, sbc_dec = {0};
bool sbc_initialized = false;
size_t sbc_codesize = 0;
uint8_t sbc_bitpool = 53;
uint8_t sbc_subbands = SBC_SB_8;
uint8_t sbc_mode = SBC_MODE_JOINT_STEREO;
bool sbc_allocation_snr = false;

// aptX encoder and decoder
struct aptx_context *aptx_enc = NULL;
struct aptx_context *aptx_dec = NULL;
bool aptx_initialized = false;

bool openal_destroy();
void sbc_encode_decode(size_t input_len);
void aptx_encode_decode(size_t input_len);
bool stop();
bool open_and_play();
void initialize_everything();
void destroy_everything();

#ifndef __EMSCRIPTEN__
void main_loop();
#endif



///////////////////////////////////

#ifdef __EMSCRIPTEN__
void 
#else
int 
#endif
process() {
    int source_state;
    /* This function is a 'main loop' for emscripten. */

    // If no file is opened, just return.
    if (!fp) {return;}
    if (feof(fp)) {

        if (close_tick != CLOSE_TICK_WAIT) {
            // Wait a bit until all buffers are played.
            close_tick++;
#ifdef __EMSCRIPTEN__
            return;
#else
            return 0;
#endif
        }
        close_tick = 0;
        fclose(fp);
        fp = NULL;
        alSourceUnqueueBuffers(source, NUM_BUFFERS, buffer);
        puts("Process: done");
        destroy_everything();
#ifdef __EMSCRIPTEN__
            return;
#else
            return 0;
#endif
    }

    alGetSourcei(source, AL_BUFFERS_PROCESSED, &buffer_queue);

    if (buffer_queue <= 0) {
#ifdef __EMSCRIPTEN__
        return;
#else
        return 2;
#endif
    }
    //sleep(1);

    //printf("trying to continue\n");
    //printf("Buffers processes: %d (out of %d)\n", buffer_queue, NUM_BUFFERS);
    buffer_queue_i = buffer_queue;
    alSourceUnqueueBuffers(source, buffer_queue, buffer);
    while(buffer_queue_i--) {
        //printf("buf queue = %d\n", buffer_queue_i);
        fr = fread(audiobuffer, 1, BUFSIZE, fp);
        if (!fr) {
            buffer_queue = buffer_queue - buffer_queue_i - 1;
            break;
        }

        if (current_codec == CODEC_SBC)
            sbc_encode_decode(fr);
        else if (current_codec == CODEC_APTX || current_codec == CODEC_APTX_HD)
            aptx_encode_decode(fr);

        //printf("UNQUEUE BUFFER ID = %d\n", buffer_queue);
        alBufferData(buffer[buffer_queue - buffer_queue_i - 1],
                        AL_FORMAT_STEREO16, audiobuffer, BUFSIZE, SAMPLE_RATE);
        alGetSourcei(source, AL_SOURCE_STATE, &source_state);
        if(source_state != AL_PLAYING)
            alSourcePlay(source);
    }
    alSourceQueueBuffers(source, buffer_queue, buffer);
#ifdef __EMSCRIPTEN__
    return;
#else
    // return WAIT_MORE here to prevent early audio output closing
    return WAIT_MORE;
#endif

}

void sbc_encode_decode(size_t input_len) {
    //puts("SBC");
    // SBC ENCODE
    size_t sbc_len, sbc_decoded_len;
    ssize_t sbc_encoded_size, sbc_decode_count, sbc_framelen;
    ssize_t sbc_pos = 0;

    size_t int_size = input_len;
    char *int_inp = audiobuffer;
    char *int_output = enc_output_audiobuffer;
    char *int_outp = enc_output_audiobuffer;
    
    //sbc_enc.bitpool = sbc_bitpool;

    while (int_size >= sbc_codesize) {
        //printf("!!!!!!!!!!!!!!!!!!\n");
        ssize_t len = sbc_encode(&sbc_enc, (const void*)int_inp, sbc_codesize,
            (void*)int_outp, (size_t)(BUFSIZE - (int_outp - int_output)),
            &sbc_encoded_size);
        //puts("Encode OK");
        if (len != sbc_codesize || sbc_encoded_size <= 0) {
            fprintf(stderr,
                "sbc_encode fail, len=%zd, encoded=%lu\n",
                len, (unsigned long) sbc_encoded_size);
            break;
        }
        int_size -= len;
        int_inp += len;
        int_outp += sbc_encoded_size;
        //printf("SBC Encoded: %zd\n", sbc_encoded_size);
    }
    
    // SBC DECODE
    sbc_framelen = sbc_decode(&sbc_dec, enc_output_audiobuffer, BUFSIZE,
                                audiobuffer, BUFSIZE, &sbc_decoded_len);
    sbc_decode_count = sbc_decoded_len;
    //printf("SBC decode Framelen = %d\n", sbc_framelen);
    
    while (sbc_framelen > 0) {
        /* push the pointer (offset) in the data forward to the next bit to be
        * decoded tell the decoder to decode up to the remaining
        * length of the buffer */
        sbc_pos += sbc_framelen;
        sbc_framelen = sbc_decode(&sbc_dec, enc_output_audiobuffer + sbc_pos, BUFSIZE - sbc_pos,
                    audiobuffer + sbc_decode_count, BUFSIZE - sbc_decode_count, &sbc_len);

        /* increase the count */
        sbc_decode_count += sbc_len;
    }

    sbc_decode_count = 0;
    sbc_pos = 0;
}

void aptx_encode_decode(size_t input_len) {
    //puts("aptX");
    // aptX ENCODE
    size_t aptx_decoded_len = 0;
    size_t aptx_encoded_size = 0;
    size_t c24 = 0;
    size_t aptx_offset = 0;

    /* Ghetto 16 to 24 bit conversion */
    for (size_t i=0; i <= input_len; i+=2) {
        int_buffer[c24] = 0u;
        int_buffer[c24+1] = (unsigned char)audiobuffer[i];
        int_buffer[c24+2] = (unsigned char)audiobuffer[i+1];
        c24+=3;
        //printf("i=%zu\n", i);
    }
    c24 -= 3;
    //c24 = 24576;
    //printf("c24 = %zu\n", c24);
    //memcpy(int_buffer, audiobuffer, input_len);

    //printf("!!!!!!!!!!!!!!!!!!\n");
    ssize_t len = aptx_encode(aptx_enc, (const unsigned char*)int_buffer, c24,
        (unsigned char*)enc_output_audiobuffer, (size_t)(BUFSIZE),
        &aptx_encoded_size);
    //puts("Encode OK");
    //printf("aptx_encode len = %zd, encoded_size = %zd\n", len, aptx_encoded_size);
    
    // aptX DECODE
    while (len > 0) {
        len = aptx_decode(
        aptx_dec,
        (const unsigned char*)(enc_output_audiobuffer+aptx_offset),
                          (aptx_encoded_size-aptx_offset),
                          int_buffer+aptx_decoded_len,
                          BUFSIZE,
                          &aptx_decoded_len);
        //printf("len = %zu, aptx_decoded_len = %zu\n", len, aptx_decoded_len);
        aptx_offset += len;
    }
    //printf("DECODED: %zu\n", aptx_offset);

    //memset(audiobuffer, 0, BUFSIZE);
    //memset(int_buffer, 0, BUFSIZE*2);
    
    /* Ghetto 24 to 16 bit conversion */
    c24 = 0;
    for (size_t i=0; i <= input_len; i+=2) {
        audiobuffer[i] = (char)int_buffer[c24+1];
        audiobuffer[i+1] = (char)int_buffer[c24+2];
        c24+=3;
        //printf("i=%zu\n", i);
    }
    //memcpy(audiobuffer, int_buffer, input_len);
}

bool openal_initialize() {
    if (openal_initialized) {
        return false;
    }

    device = alcOpenDevice(NULL);
    if (device == NULL) {
    // Handle Exception
        perror("OpenAL: device is null");
    }

    //Create a context
    context=alcCreateContext(device,NULL);
    if (!context) {
        perror("OpenAL: context is null");
    }
    
    //Set active context
    if (!alcMakeContextCurrent(context)) {
        perror("OpenAL: can't switch context");
    }
    // Clear Error Code
    alGetError();

    alGenSources((ALuint)1, &source);
    alGenBuffers((ALuint)NUM_BUFFERS, buffer);

    openal_initialized = true;
    return true;
}

bool openal_destroy() {
    if (!openal_initialized) {
        return false;
    }
    puts("OpenAL: closing");
    alDeleteBuffers(NUM_BUFFERS, buffer);
    alDeleteSources(1, &source);
    device = alcGetContextsDevice(context);
    alcMakeContextCurrent(NULL);
    alcDestroyContext(context);
    alcCloseDevice(device);

    openal_initialized = false;
    return true;
}

bool sbc_initialize() {
    if (sbc_initialized) {
        puts("SBC already initialized, not initializing.");
        return false;
    }
    memset(&sbc_enc, 0, sizeof(sbc_t));
    memset(&sbc_dec, 0, sizeof(sbc_t));
    sbc_init(&sbc_enc, 0L);
    sbc_init(&sbc_dec, 0L);

    sbc_enc.frequency = SBC_SAMPLE_RATE;
    sbc_enc.subbands = sbc_subbands;
    //printf("(int) SBC SUBBANDS: %x\n", sbc_enc.subbands);
    sbc_enc.mode = sbc_mode;
    //printf("(int) SBC MODE: %x\n", sbc_enc.mode);
    sbc_enc.endian = SBC_LE;
    sbc_enc.allocation = (sbc_allocation_snr ? SBC_AM_SNR : SBC_AM_LOUDNESS);
    sbc_enc.blocks = SBC_BLK_16;
    sbc_enc.bitpool = sbc_bitpool ? sbc_bitpool : 53;
    //printf("(int) SBC BITPOOL: %u\n", sbc_enc.bitpool);

    sbc_dec.endian = SBC_LE;

    sbc_codesize = sbc_get_codesize(&sbc_enc);
    //printf("SBC Codesize = %zd\n", sbc_codesize);

    sbc_initialized = true;
    puts("SBC initialized.");
    return true;
}

bool sbc_destroy() {
    if (!sbc_initialized) {
        puts("SBC is not initialized, not destroying.");
        return false;
    }
    sbc_finish(&sbc_enc);
    sbc_finish(&sbc_dec);

    sbc_initialized = false;
    puts("SBC destroyed.");
    return true;
}

bool aptx_initialize(bool hd) {
    if (aptx_initialized) {
        return false;
    }

    aptx_enc = aptx_init(hd);
    aptx_dec = aptx_init(hd);

    aptx_initialized = true;
    return true;
}

bool aptx_destroy() {
    if (!aptx_initialized) {
        return false;
    }
    aptx_finish(aptx_enc);
    aptx_finish(aptx_dec);
    aptx_enc = NULL;
    aptx_dec = NULL;
    aptx_initialized = false;
    return true;
}

void sbc_set_bitpool(uint8_t bitpool) {
    printf("SBC Bitpool: %d\n", bitpool);
    //sbc_destroy();
    sbc_bitpool = bitpool;
    sbc_enc.bitpool = bitpool;
    //sbc_initialize();
}

void sbc_set_mode(uint8_t mode) {
    if (mode >= 4) {
        puts("Mode value unknown!");
        return;
    }
    sbc_destroy();
    sbc_mode = mode;
    printf("SBC: set %s\n", mode_names[mode]);
    sbc_initialize();
}

void sbc_set_subbands(uint8_t subbands) {
    if (subbands != 8 && subbands != 4) {
        puts("Subband value unknown!");
        return;
    }
    sbc_destroy();
    sbc_subbands = (subbands == 8 ? SBC_SB_8 : SBC_SB_4);
    printf("SBC: set Subbands to %u\n", subbands);
    sbc_initialize();
}

void sbc_set_allocation_snr(bool snr) {
    sbc_destroy();
    if (snr) snr = 1; // force bool
    sbc_allocation_snr = (snr ? SBC_AM_SNR : SBC_AM_LOUDNESS);
    printf("SBC: set Allocation_SNR to %u\n", snr);
    sbc_initialize();
}

bool set_codec(uint8_t codec) {
    if (current_codec == codec) {
        return false;
    }

    if ((current_codec != CODEC_APTX && codec == CODEC_APTX) ||
        (current_codec != CODEC_APTX_HD && codec == CODEC_APTX_HD)
    ) {
        // from aptX/HD to SBC
        puts("Reinitializing aptX");
        aptx_destroy();
        aptx_initialize(codec == CODEC_APTX_HD);
    }
    printf("Switching CODEC from %s to %s\n",
           codec_names[current_codec],
           codec_names[codec]
          );
    current_codec = codec;
    return true;
}

bool stop() {
    if (fp) {
        fclose(fp);
        fp = NULL;
    }
    destroy_everything();
    return true;
}

bool open_and_play(char *filename) {
    if (fp) {
        fclose(fp);
        fp = NULL;
    }

    destroy_everything();
    initialize_everything();

    fp = fopen(filename, "r");
    if (!fp) {
        perror("Can't open audio file!");
        return false;
    }

    // fill buffers
    for (int i=0; i<NUM_BUFFERS; i++) {
        fr = fread(audiobuffer, 1, BUFSIZE, fp);
        if (!fr) {
            perror("File is too small!");
            return false;
        }

        if (current_codec == CODEC_SBC)
            sbc_encode_decode(fr);
        else if (current_codec == CODEC_APTX || current_codec == CODEC_APTX_HD)
            aptx_encode_decode(fr);

        alBufferData(buffer[i], AL_FORMAT_STEREO16, audiobuffer, BUFSIZE, SAMPLE_RATE);
    }

    //alSourcei(source, AL_BUFFER, buffer[0]);
    alSourceQueueBuffers(source, NUM_BUFFERS, buffer);
    alSourcePlay(source);
    return true;
}

#ifndef __EMSCRIPTEN__
void main_loop() {
    open_and_play("sine.pcm");

    while (1) {
        int process_output = process();
        if (process_output == WAIT_MORE) {
            usleep(200000);
        }
        else if (!process_output) {
            break;
        }
    }
}
#endif

void initialize_everything() {
    if (openal_initialized) {
        puts("OpenAL: already initialized");
    }
    else {
        openal_initialize();
    }

    if (sbc_initialized) {
        puts("SBC: already initialized");
    }
    else {
        sbc_initialize();
    }
    sbc_set_bitpool(sbc_bitpool);
    sbc_set_mode(sbc_mode);
    sbc_set_allocation_snr(sbc_allocation_snr);

    if (aptx_initialized) {
        puts("aptX: already initialized");
    }
    else {
        aptx_initialize(current_codec == CODEC_APTX_HD);
    }
}

void destroy_everything() {
    sbc_destroy();
    aptx_destroy();
    openal_destroy();
}

int main() {
    //initialize_everything();

#ifndef __EMSCRIPTEN__
    main_loop();
    puts("This is the end");
    openal_destroy();
#else
    puts("Bluetooth SBC/aptX audio encoder by ValdikSS");
    puts("Uses GNU GPL BlueZ SBC library copyright (c) 2004-2019 Marcel Holtmann et al");
    puts("Uses GNU LGPL libopenaptx aptX library by pali");
    puts("Compiled from C with emscripten");
    emscripten_set_main_loop(process, EMSCRIPTEN_FPS, 1);
#endif
}